/////////////////////////////////////////////////////////////////////////////
// Charset to Code Page static table.

// ANSI_CHARSET            0     1252
// SHIFTJIS_CHARSET      128      932
// HANGEUL_CHARSET       129      949
// GB2312_CHARSET        134      936
// CHINESEBIG5_CHARSET   136      950
// GREEK_CHARSET         161     1253
// TURKISH_CHARSET       162     1254
// HEBREW_CHARSET        177     1255
// ARABIC_CHARSET        178     1256
// BALTIC_CHARSET        186     1257
// RUSSIAN_CHARSET       204     1251
// THAI_CHARSET          222      874
// EASTEUROPE_CHARSET    238     1250

/////////////////////////////////////////////////////////////////////////////
// Operating system identification.

#define OS_WIN95            0x03
#define OS_WIN32s           0x02
#define OS_NT               0x00

g_os_version = (GetVersion() >> 30) & 0x0003;

/////////////////////////////////////////////////////////////////////////////
// Code Page enumeration building data tables.

#define char_types(w)            (*(pbCharTypes[BYTE(w>>8)] + BYTE(w)))
#define set_char_types(w, bType) (*(pbCharTypes[BYTE(w>>8)] + BYTE(w))  = bType)
#define add_char_types(w, bType) (*(pbCharTypes[BYTE(w>>8)] + BYTE(w)) |= bType)

#define SORT_KEY_SEPARATOR  0x01

#define DIGIT_CHAR      	0x01
#define LETTER_CHAR     	0x02
#define LETTER_IMBED    	0x04
#define DIGIT_IMBED     	0x08
#define SPACE_CHAR      	0x10
#define UPPER_CASE_CHAR 	0x20
#define ACCENTED_CHAR   	0x40
#define CHAR_DEFINED        0x80
#define WORD_TYPE           (LETTER_CHAR | DIGIT_CHAR)

BOOL CALLBACK CodePageEnumProc(LPSTR lpCodePageString)
{
	BYTE   bSection;
	BYTE   szChars[2];
	LCID   lcid;
	int    i, j, nCount, nFinal;
	WCHAR  wChars;
	WORD   wCharType1, wCharType2, wCharType3;
	CP     wCP;
	CPINFO CPInfo;

	wCP  = atoi(lpCodePageString);
	lcid = GetLocaleFromCP(wCP);
	
	if (!GetCPInfo(wCP, &CPInfo))
		return TRUE;

	if (nFinal = (CPInfo.MaxCharSize == 1) ?  0 : 255)			// one pass if no lead bytes (MaxCharSize = 1)
	{
		g_lastCP = wCP;

		memset(bLeadBytes, 0, sizeof(bLeadBytes));
	
		for (i = 0; i < MAX_LEADBYTES; i += 2)
		{
			if (!CPInfo.LeadByte[i] && !CPInfo.LeadByte[i+1])
				break;											// end of lead byte ranges
		
			for (j = CPInfo.LeadByte[i]; j <= CPInfo.LeadByte[i+1]; j++)
				bLeadBytes[j] = TRUE;							// mark as valid lead byte
		}
	}

	for (i = 0; i <= nFinal; i++)								// thumb thru all potential lead bytes
	{
		if (!i || bLeadBytes[i])								// lead bytes OR chars 0x00 - 0xff
		{
			for (j = 0; j < 256; j++)
			{													  
				nCount = 0;
				if (i)
					szChars[nCount++] = i;						// create leadbyte/char pairs
				szChars[nCount++] = j;

				if (MultiByteToWideChar(wCP, MB_ERR_INVALID_CHARS, (PSTR)szChars, nCount, (PWSTR)&wChars, 1) != 1)
					continue;									// not valid UNICODE character

				bSection = HIBYTE(wChars);

				if (pbCharTypes[bSection] == bDefaultTable)		// UNICODE section not accessed yet
				{
					pbCharTypes[bSection] = new BYTE[256];

					memset(pbCharTypes[bSection], 0, 256 * sizeof(BYTE));
				}
																// already processed this UNICODE char			
				else if (char_types(wChars))
					continue;
											 
				GetStringTypeA(lcid, CT_CTYPE1, (PSTR)szChars, i ? 2 : 1, &wCharType1);
				GetStringTypeA(lcid, CT_CTYPE2, (PSTR)szChars, i ? 2 : 1, &wCharType2);
				GetStringTypeA(lcid, CT_CTYPE3, (PSTR)szChars, i ? 2 : 1, &wCharType3);

				set_char_types(wChars, CHAR_DEFINED);

				if (wCharType1 & C1_ALPHA)						// process characters
					add_char_types(wChars, LETTER_CHAR);
	
				if (wCharType1 & C1_SPACE)
					add_char_types(wChars, SPACE_CHAR);			// mark space characters

				if ((wCharType1 & C1_DIGIT) || (wCharType2 == C2_EUROPENUMBER) || (wCharType2 == C2_ARABICNUMBER))
					add_char_types(wChars, DIGIT_CHAR);			// mark number characters

				if (wCharType3 & C3_LEXICAL)
					add_char_types(wChars, LETTER_IMBED);		// mark letter embedded separators

				if (wCharType2 == C2_COMMONSEPARATOR || wCharType2 == C2_EUROPESEPARATOR)
					add_char_types(wChars, DIGIT_IMBED);		// mark number embedded separators				
			}
		}
	}

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// Unicode word lex routine.

// fTokenizeSpaces flags:
                                        // removes single space between tokens (ftsrch mode).
#define TOKENIZE_SPACES     0x01        // treat spans of spaces as separate token (WinHelp mode).
#define REMOVE_SPACE_CHARS  0x02        // remove space characters from punctuation tokens.
#define STARTING_IMBEDS     0x04        // allow imbeds as starting character in token (for edit input).

int FAR PASCAL WordBreakW (LPWSTR *ppwText, LPINT pcwText, LPWSTR *paToken, LPWSTR *paTokenEnd,
						   LPBYTE paType, PUINT paHash, int cwTokens, UINT fTokenizeSpaces)
{
	BYTE    bCharType, bPrevType, bFirstCharType;
	UINT	wHash;
	WORD    wPunc, cwTokensOut = 0;
	WCHAR	wChar, wChar2;
	LPWSTR  pwPos, pwLimit, pwTokenStart;

	pwPos  = *ppwText;											// position WCHAR pointer to beginning of text
	wChar  = *pwPos;											// get first UNICODE character 
	pwLimit = pwPos + *pcwText;									// end of UNICODE text

	FOREVER_
	{															// token hash value init
		wHash = 0;

		if (pwPos == pwLimit) 									// have reached end of UNCODE text
			break;

		bFirstCharType = (char_types(wChar) & WORD_TYPE);
		bPrevType = 0;

        if (!bFirstCharType && (fTokenizeSpaces & STARTING_IMBEDS))
        {
            bCharType = char_types(wChar);
            if (bCharType & LETTER_IMBED)
            {
                if (pwPos+1 == pwLimit || char_types(*(pwPos+1)) & LETTER_CHAR)
                {
                    bFirstCharType = TRUE;
                    bPrevType |= LETTER_CHAR;
                }
            }

            if (bCharType & DIGIT_IMBED)
            {
                if (pwPos+1 == pwLimit || char_types(*(pwPos+1)) & DIGIT_CHAR)
                {
                    bFirstCharType = TRUE;
                    bPrevType |= DIGIT_CHAR;
                }
            }
        }

		if (bFirstCharType)										// current WCHAR is letter or number
		{				 
			pwTokenStart = pwPos;								// save pointer to beginning of token
			wHash = 0;									    	// seed hash value

			FOREVER_
			{
				do
				{
					wChar = *pwPos;								// current UNICODE character
					bCharType = char_types(wChar);

					if ((bCharType & WORD_TYPE) || 

					   ((bCharType & LETTER_IMBED) &&			// changed to allow C3_LEXICAL (letter
					   ((bPrevType & LETTER_CHAR))) ||			// ... imbed) to be suffix

					   ((bCharType & DIGIT_IMBED) &&
					    (bPrevType & DIGIT_CHAR)  &&
						(pwPos+1 == pwLimit || char_types(*(pwPos+1)) & DIGIT_CHAR || (fTokenizeSpaces & STARTING_IMBEDS))))

						{
						wHash = HashAChar(wHash, wChar);	   	// token continues: letter, number, or
						bPrevType = bCharType;					// ... surrounded embedded character
						}

					else
						break;									// else token complete
				}								 
				while (++pwPos != pwLimit);						// until end of UNICODE text

				if (!cwTokens)											
					cwTokensOut++;								// just count number of tokens needed

				else
				{
					if (paToken)
 						paToken[cwTokensOut] = pwTokenStart;	// token start pointer
		
					if (paTokenEnd)
 						paTokenEnd[cwTokensOut] = pwPos;		// token end pointer

					if (paHash)
						paHash[cwTokensOut] = wHash;			// token hash value

 					if (paType)
						paType[cwTokensOut] = bFirstCharType;	// mark token as word (chars/digits)

					if (++cwTokensOut >= cwTokens)				// no more token pointer space
					{
						*pcwText -= (pwPos - *ppwText);			// update UNICODE character count
						*ppwText = pwPos;						// update WCHAR text starting pointer
						return(cwTokensOut);					// return token count
					}
				}
                                                                // remove all spans of space characters
				if ((fTokenizeSpaces & REMOVE_SPACE_CHARS) && pwPos != pwLimit)		
				{                                               
					while (pwPos != pwLimit && (char_types(*pwPos) & SPACE_CHAR))
						pwPos++;
                
                    if (pwPos == pwLimit)
                        break;

					pwTokenStart = pwPos;
					wChar = *pwPos;
					wHash = 0;

					if (!(char_types(wChar) & WORD_TYPE))		// lexing into non-space punctuation
						break;
				}

				else if (!(fTokenizeSpaces & TOKENIZE_SPACES) && pwPos != pwLimit &&
				    wChar == UNICODE_SPACE_CHAR && (pwPos+1) != pwLimit &&
					char_types(wChar2 = *(pwPos+1)) & WORD_TYPE)
				{	
					pwTokenStart = ++pwPos;						// if "fTokenizeSpaces" is FALSE, then
					wHash = 0;  							    // ... remove single space between words					continue;
				}												// ... as a token	
				else 
					break;
			}
		}
	
		if (pwPos == pwLimit) break; 					        // ... at end of provided WCHAR text
	
		pwTokenStart = pwPos;									// save pointer to beginning of token
		wHash = 0;										     	// seed hash value
		wPunc = wChar;											// punctuation type (space vs. non-space)
	
		do
		{
			wChar = *pwPos;										// current UNICODE character 

			if (fTokenizeSpaces & TOKENIZE_SPACES)				// "fTokenizeSpaces" option for WinHelp
				if ((wPunc == UNICODE_SPACE_CHAR && wChar != UNICODE_SPACE_CHAR) ||
					(wPunc != UNICODE_SPACE_CHAR && wChar == UNICODE_SPACE_CHAR))
					break;										// tokenize spans of spaces -OR- non-spaces
		
			bCharType = char_types(wChar);
			if (!(bCharType & WORD_TYPE) || !wChar)
				{
				if (!(fTokenizeSpaces & REMOVE_SPACE_CHARS) || !(bCharType & SPACE_CHAR))
					wHash = HashAChar(wHash, wChar);			// punctuation token continues: not letter/number
				}
			else
				break;
		}
		while (++pwPos != pwLimit);								// until end of UNICODE text


		if (pwPos != pwLimit || pwTokenStart+1 != pwLimit || wChar)
		{													   	// discard empty final token
			LPWSTR pw, pwNew = pwPos;

			if (fTokenizeSpaces & REMOVE_SPACE_CHARS)			// remove spans of space chars
			{
				for (pw = pwNew = pwTokenStart; pw < pwPos;	pw++)	
			 		if (!(char_types(*pw) & SPACE_CHAR))
						*pwNew++ = *pw;
			}

			if (pwNew != pwTokenStart)
			{
				if (!cwTokens)											
					cwTokensOut++;								// just count number of tokens needed

				else
				{
					if (paToken)
	 					paToken[cwTokensOut] = pwTokenStart;	// Token start pointer

					if (paTokenEnd)
	 					paTokenEnd[cwTokensOut] = pwNew;		// Token end pointer

					if (paHash)
						paHash[cwTokensOut] = wHash;			// Token hash value

	 				if (paType)
						paType[cwTokensOut] = 0;				// mark token as punctuation

					if (++cwTokensOut >= cwTokens)
					{
						*pcwText -= (pwPos - *ppwText);			// update UNICODE character count
						*ppwText = pwPos;						// update WCHAR text starting pointer
						return(cwTokensOut);	 				// return token count
					}
				}
			}
		}
	}

	if (cwTokens)
	{
		*pcwText -= (pwPos - *ppwText);							// update UNICODE character count
		*ppwText = pwPos;										// update WCHAR text starting pointer
	}

	return cwTokensOut;											// return token count
}			

/////////////////////////////////////////////////////////////////////////////
// Sort Key conversion to format usable for prefix/suffix/substring tests.

UINT SortKeyText(PWCHAR pwText, UINT cwText, PWCHAR pwOut, UINT cwOut)
{                                                       // convert sort key into text order based representation
    UINT   nChar = 0;                                   // ... for IsAPrefix and IsASuffix type text operations.
                                                        // ... Work field is built up as individual alpha sort
	PCHAR  pWork, pAlpha, pBuild, pSeparator;           // ... weights immediately followed by their diacritic
	PWCHAR pwTemp;                                      // ... and case weight bytes: [WORD][BYTE-BYTE] per char.
                                                        // ... Non-existent diacritic and case weights are filled
    PWCHAR pwEnd  = pwText + cwText;                    // ... out with SORT_KEY_SEPARATOR (lowest sort value).
	PCHAR  pStart = (PCHAR)pwOut;
                                                        
    if (!cwText)                                        // NOTE: Uses NLSAPI V6.0 diacritic/case weight algorithm
        return 0;                                       // ... removing non-significant weights from left-to-right.
                                                        // ... V6.1 removes non-significant weights right-to-left.
	while (HIBYTE(*pwText) != SORT_KEY_SEPARATOR)                   
    {                                                               // search for first weight separator
		pwOut[nChar++] = *pwText++;               
        pwOut[nChar++] = MAKEWORD(SORT_KEY_SEPARATOR, SORT_KEY_SEPARATOR);
        if (nChar >= cwOut)
            return 0;
    }
    pwOut[nChar] = 0;                                               // terminating NULL for "strstr"

	for (pWork = (PCHAR)pwText + 1; pWork < (PCHAR)pwEnd; pWork++)  // skip diacritic separator
		if (*pWork == SORT_KEY_SEPARATOR)                           // find alpha weights separator
			break;

    pBuild = PCHAR(pwOut + nChar) - 1;
    for (pAlpha = pWork - 1; pAlpha > (PCHAR)pwText && pBuild > pStart; pAlpha--)
    {
        *pBuild = *pAlpha;                                          // fill out diacritic weights
        pBuild -= 4;                                                // next diacritic, moving from end to start
    }

    pSeparator = pWork++;
    for ( ; pWork < (PCHAR)pwEnd; pWork++)                          // skip case separator
		if (*pWork == SORT_KEY_SEPARATOR)                           // find special weights separator
			break;

    pBuild = PCHAR(pwOut + nChar) - 2;
    for (pAlpha = pWork - 1; pAlpha > pSeparator && pBuild > pStart; pAlpha--)
    {
        *pBuild = *pAlpha;                                          // fill out case weights
        pBuild -= 4;                                                // next case, moving from end to start
    }

	return nChar;
}

////////////////////////////////////////////////////////////////////////////////////////
// Lower case input (pwL) character matches lower case, upper case, and diacritics.
// Upper case input (pwL) character only matches upper case comparison character in pwR.
// Diacritic input (pwL) character only matches diacritic comparison character in pwR.

BOOL HasASuffix(PWCHAR pwL, UINT cwL, PWCHAR pwR, UINT cwR)
{
    BYTE   bChar;

    if (cwL > cwR) return FALSE;

    pwR += cwR - cwL;

    for (cwL >>= 1; cwL--; pwL++, pwR++)
    {
        if (*pwL++ != *pwR++)
        	return FALSE;

        bChar = HIBYTE(*pwL);
        if (bChar != SORT_KEY_SEPARATOR && bChar != HIBYTE(*pwR))
            return FALSE;
    
        bChar = LOBYTE(*pwL);
        if (bChar != SORT_KEY_SEPARATOR && bChar != LOBYTE(*pwR))
            return FALSE;
    }

    return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// Sort key generation from Unicode.

int FAR PASCAL LCSortKeyW(LCID lcid, LPCWSTR pwSource, int cwSource, LPWSTR pwDest, int cwDest)
{
	int cb, nRet;

	if (g_os_version == OS_NT)
		nRet = LCMapStringW(lcid, LCMAP_SORTKEY, pwSource, cwSource, pwDest+1, (cwDest-1)<<1) >> 1;

    else
	{
		cwSource <<= 1;											// 1 WC can generate 2 bytes of MB

		if (cwSource > cbSource)
		{
			cbSource = 0;
			if (pbSource) delete pbSource;
			if (pbSource = new BYTE[cwSource]) cbSource = cwSource;
		}

		if (!pbSource) return 0;								// error return

		cb = WideCharToMultiByte(GetCPFromLocale(lcid), 0, pwSource, cwSource>>1, (PSTR)pbSource, cbSource, NULL, NULL);

		nRet = LCMapStringA(lcid, LCMAP_SORTKEY, (PSTR)pbSource, cb, (PSTR)(pwDest+1), (cwDest-1)<<1) >> 1;
	}

	if (nRet)
	{
		nRet++;

		if (cwDest && pwDest)									// set a sort keys prefix so tokens group first by
		{
			BYTE bCharType = char_types(*pwSource);

			*pwDest = ~(bCharType & WORD_TYPE);					// ... alphabetics, then numerics, then punctuation
		}
	}

	return nRet;
}


/////////////////////////////////////////////////////////////////////////////
// Routine to skip multiple DBCS characters.

LPSTR FAR PASCAL CharNextMult(CP wCP, LPCSTR str, int n)
{
	int i, j;

	if (wCP != g_lastCP)								// we are processing a new CP, so
	{													// ... set up our lead byte tables
		CPINFO CPInfo;

		if (!GetCPInfo(wCP, &CPInfo))
			return (LPSTR)str + n;						// error return, let's make a guess

		g_lastCP = wCP;

		memset(bLeadBytes, 0, sizeof(bLeadBytes));	   	// establish lead bytes

		for (i = 0; i < MAX_LEADBYTES; i += 2)
		{
			if (!CPInfo.LeadByte[i] && !CPInfo.LeadByte[i+1])
				break;									// end of lead byte ranges

			for (j = CPInfo.LeadByte[i]; j <= CPInfo.LeadByte[i+1]; j++)
				bLeadBytes[j] = TRUE;					// mark as valid lead byte
		}
	}

	for (i = 0; i < n; i++, str++)
		if (bLeadBytes[*str])
			str++;

	return (LPSTR)str;
}


